\switch +gump +compilerpasses 

%%%
%%% Authors:
%%%   Andreas Sundstrom (andreas@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzen, 1999
%%%   Andreas Sundstrom, 1999
%%%
%%% Last change:
%%%   $Date: 2000/03/23 12:40:00 $ by $Author: andreas $
%%%   $Revision: 1.20 $
%%%
%%% See the file "LICENSE" or
%%%    http://mozart.ps.uni-sb.de/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
% Loads a mozart source file, makes lexical analysis and returns a abstract description of it's contents
*/

functor
import
   GumpScanner('class')
   GumpParser('class')
   Property(get)
   Open(file)
   OS(getEnv stat)
   System(showInfo)% show:Show)
   SymTable(symTable:SymbolTable)
export
   parseFile: ParseFile
   parseVirtualString: ParseVirtualString
define
   local
      fun {Wrapper A Checker WrapChar}
	 CharList = {Atom.toString A}
      in
	 if {Not {Char.isAlpha CharList.1}} then
	    {String.toAtom WrapChar|{List.append CharList [WrapChar]}}
	 elseif {Checker CharList.1} then
	    {String.toAtom WrapChar|{List.append CharList [WrapChar]}}
	 else
	    A
	 end
      end
      proc {FlattenListArgs Xs ?Ys ?Tail}
	 case Xs 
	 of fAtom(nil) then
	    Ys = nil
	    Tail = false
	 [] fRecord(fAtom('|') [Y Xr]) then Yr in
	    Ys = Y|Yr
	    {FlattenListArgs Xr ?Yr ?Tail}
	 [] fVar(V) then 
	    Ys = fVar(V)|nil
	    Tail = true
	 [] '_' then
	    Ys = '_'|nil
	    Tail = true
	 end
      end
   in
      fun {ChangeToAtom X}
	 case X
	 of fVar(V) then
	    V
	 [] fAtom(A) then
	    {Wrapper A Char.isUpper &'}
	 [] fInt(I) then
	    {String.toAtom {Int.toString I}}
	 [] fEscape(V) then
	    {VirtualString.toAtom '!'#{ChangeToAtom V}}
	 [] fEq(LH RH _) then
	    {VirtualString.toAtom {ChangeToAtom LH}#' = '#{ChangeToAtom RH}}
	 [] fRecord(L Args) then Temp Temp2 in
	    case L
	    of fAtom('#') then
	       Temp = {List.foldL Args fun {$ X Y}
					  {ChangeToAtom X}#'*'#{ChangeToAtom Y}
				       end ''}
	       Temp2 = {VirtualString.toString Temp}.2
	       {VirtualString.toAtom {List.map Temp2
				      fun {$ X}
					 if X == &* then
					    &#
					 else
					    X
					 end
				      end}}
	    [] fAtom('|') then FlatList HasTail in
	       {FlattenListArgs X ?FlatList ?HasTail}
	       if HasTail then
		  Temp = {List.foldL FlatList
			  fun {$ X Y}
			     {ChangeToAtom X}#'|'#{ChangeToAtom Y}
			  end ''}
		  Temp2 = {VirtualString.toString Temp}.2
		  {VirtualString.toAtom Temp2}
	       else
		  Temp = {List.foldL FlatList
			  fun {$ X Y}
			     {ChangeToAtom X}#' '#{ChangeToAtom Y}
			  end ''}
		  Temp2 = {VirtualString.toString Temp}.2
		  {VirtualString.toAtom '['#Temp2#']'}
	       end
	    else
	       Temp = {List.foldL Args fun {$ X Y}
					  {ChangeToAtom X}#' '#{ChangeToAtom Y}
				       end ''}
	       Temp2 = {VirtualString.toString Temp}.2
	       {VirtualString.toAtom {ChangeToAtom L}#'('#Temp2#')'}
	    end
	 [] fColon(X Y) then Temp in
	    Temp = {ChangeToAtom X}#':'#{ChangeToAtom Y}
	    {VirtualString.toAtom Temp}
	 [] fOpApply('.' [X Y]) then Temp in
	    Temp = {ChangeToAtom X}#'.'#{ChangeToAtom Y}
	    {VirtualString.toAtom Temp}
	 else  % _ or $
	    X  
	 end
      end
   end
	 
   \insert OzScanner.ozg
   \insert OzParser.ozg

   local
      InfoObject = {Cell.new unit}
      proc {Info Msg}
	 {{Cell.access InfoObject} Msg}
      end
      
      proc {FunctorDescriptors Xs}
	 case Xs
	 of X|Xr then
	    case X
	    of fImport(Y) then
	       {Info add('import' Y)}
	    [] fExport(Y) then
	       {Info add('export' Y)}
	    [] fDefine(Y) then
	       {Info add('define' Y)}
	    [] fRequire(Y) then
	       {Info add('require' Y)}
	    [] fPrepare(Y) then
	       {Info add('prepare' Y)}
	    end
	    {FunctorDescriptors Xr}
	 [] nil then
	    skip
	 end
      end
      fun {Union Xs Ys}
	 case Xs
	 of X|Xr then
	    if {List.member X Ys} then
	       {Union Xr Ys}
	    else
	       {Union Xr X|Ys}
	    end
	 [] nil then
	    Ys
	 end
      end
      class InfoClass
	 feat
	    symTable
	    infoTable
	    functorCode
	    result
	    returnAll
	 attr
	    currentEnvironment
	    environmentStack
	    functorStack
	    inDeclaration
	    leftHand
	    'export'
	    'import'
	    'define'
	    'require'
	    'prepare'
	    comment
	    functorName
	    inPrepare
	    refsInPrepare
	    importrefs
	    
	 meth init(returnAll: All <= false)
	    self.symTable = {New SymbolTable init}
	    self.infoTable = {Dictionary.new}
	    currentEnvironment <- {Name.new}
	    environmentStack <- nil
	    inDeclaration <- 0
	    leftHand <- false      
	    'export' <- nil
	    'import' <- nil
	    'define' <- nil
	    'require' <- nil
	    'prepare' <- nil
	    functorStack <- nil
	    inPrepare <- false
	    importrefs <- nil
	    refsInPrepare <- nil
	    self.returnAll = All
	 end
	 meth setFunctorCode(FC)
	    self.functorCode = FC
	 end
	 meth pushFunctor()
	    functorStack <- 'functor'('export':@'export' 'import':@'import'
				      'define':@'define' 'require':@'require'
				      'prepare':@'prepare' comment:@comment
				      name:@functorName)|@functorStack
	    'export' <- nil
	    'import' <- nil
	    'define' <- nil
	    'require' <- nil
	    'prepare' <- nil
	    environmentStack <- @currentEnvironment|@environmentStack
	    currentEnvironment <- {Name.new}
	 end
	 meth popFunctor() F Fr in
	    F|Fr = @functorStack
	    'export' <- F.'export'
	    'import' <- F.'import'
	    'define' <- F.'define'
	    'require' <- F.'require'
	    'prepare' <- F.'prepare'
	    comment <- F.comment
	    functorName <- F.name
	    functorStack <- Fr
	    InfoClass, endScope
	 end	    
	 meth add(Description Info)
	    if @Description == nil then
	       Description <- Info
	    else
	       raise
		  analyze_error('Two functor descriptions of the same type')
	       end
	    end
	 end
	 meth setComment(C)
	    comment <- C
	 end
	 meth setFunctorName(FN)
	    functorName <- FN
	 end
	 meth analyze(Code $)
	    InfoClass, analyzeDefine(Code)
	    self.result
	 end
	 meth analyzeFunctor($)
	    proc {Filter Xs Ys ?Zs}
	       case Xs
	       of X|Xr then
		  if {List.member X Ys} then
		     {Filter Xr Ys ?Zs}
		  else Zr in
		     Zs = X|Zr
		     {Filter Xr Ys ?Zr}
		  end
	       [] nil then
		  Zs = nil
	       end
	    end
	    proc {GetAllRefs Ks D}
	       case Ks
	       of K|Kr then Id Info in
		  {self.symTable getScope(name:K info:Info id:Id)}
		  if Info \= empty andthen {Not {Dictionary.member D Id}} then 
		     {Dictionary.put D Id Info}
		  end
		  {GetAllRefs Kr D}
	       [] nil then
		  skip
	       end
	    end
		  
	    %% Follows inheritance path
	    proc {BuildRefList Xs Dict Added ?Ys}
	       case Xs
	       of X|Xr then Y Yr Info in
		  Ys = Y|Yr
		  Info = {Dictionary.get Dict X}
		  case Info
		  of 'class'(...) then WorkList NewRefs in
		     NewRefs = Info.refs
		     Y = X#'class'(name:Info.name inherits:Info.inherits
				   features:Info.features id:X
				   methods:Info.methods comment:Info.comment)
		     WorkList = {Union {Filter NewRefs X|Added} Xr}
		     {BuildRefList WorkList Dict X|Added Yr}
		  else
		     Y = X#Info
		     {BuildRefList Xr Dict X|Added Yr}
		  end
	       [] nil then
		  Ys = nil
	       end
	    end
	    %% Return a list with references in Xs but not in RefsInPrepare
	    proc {GetRefsInDefine Xs RefsInPrepare ?Ys}
	       case Xs
	       of Ref#_|Xr then Yr in
		  if {List.member Ref RefsInPrepare} then
		     Ys = Yr
		  else
		     Ys = Ref|Yr
		  end
		  {GetRefsInDefine Xr RefsInPrepare Yr}
	       [] nil then
		  Ys = nil
	       end
	    end
	    proc {AddToRefList Xs RefListIn ?Ys ?Zs}
	       case Xs
	       of X|Xr then Yr Zr in
		  Ys = X.id#X|Yr
		  Zs = X.id|Zr
		  {AddToRefList Xr RefListIn ?Yr ?Zr}
	       [] nil then
		  Ys = RefListIn
		  Zs = nil
	       end
	    end
	    proc {RemoveDuplicates Xs Ys}
	       case Xs
	       of X|Xr then
		  if {List.member X Xr} then
		     {RemoveDuplicates Xr Ys}
		  else Yr in
		     Ys = X|Yr
		     {RemoveDuplicates Xr Yr}
		  end
	       [] nil then
		  Ys = nil
	       end
	    end
	    ExportItems Refs Refs1 RefList RefsInDefine
	    RefList2 RefList3 RefList4 RefRecord
	    Imports Requires Exports Prepares
	 in
	    InfoClass, analyzeExport(@'export')
	    InfoClass, analyzeImport(fImport @'import')
	    InfoClass, analyzeImport(fRequire @'require')
	    inPrepare <- true
	    InfoClass, analyzeDefine(@'prepare')	    
	    inPrepare <- false
	    InfoClass, analyzeDefine(@'define')
	    InfoClass, exportItems(@'export' ?ExportItems ?Refs1)
	    Refs = {RemoveDuplicates Refs1}
	    if self.returnAll then
	       Keys D = {Dictionary.clone self.infoTable}
	    in
	       {self.symTable getKeys(keys:Keys)}
	       {GetAllRefs Keys D}
	       {BuildRefList {Dictionary.keys D} D nil RefList}
	    else
	       {BuildRefList Refs self.infoTable nil RefList}
	    end
	    {GetRefsInDefine RefList {Append @refsInPrepare @importrefs}
	     ?RefsInDefine}
	    {AddToRefList @'import' RefList ?RefList2 ?Imports}
	    {AddToRefList @'require' RefList2 ?RefList3 ?Requires}
	    {AddToRefList ExportItems RefList3 ?RefList4 ?Exports}
	    RefRecord = {List.toRecord refs RefList4}
	    Prepares = {List.filter @refsInPrepare
			fun {$ R} {Value.hasFeature RefRecord R} end}
	    'functor'(name:@functorName 'import':Imports 'export':Exports
		      'require':Requires 'define':RefsInDefine id:_
		      'prepare':Prepares comment:@comment
		      refs:RefRecord)
	 end
	 meth analyzeExport(Xs)
	    case Xs
	    of fExportItem(_#fVar(V))|Xr then
	       {self.symTable startScope(name:{ChangeToAtom fVar(V)} info:empty
					 env:@currentEnvironment)}
	       InfoClass, analyzeExport(Xr)
	    [] fExportItem(fVar(V))|Xr then
	       {self.symTable startScope(name:{ChangeToAtom fVar(V)} info:empty
					 env:@currentEnvironment)}
	       InfoClass, analyzeExport(Xr)
	    [] nil then
	       skip
	    end
	 end
	 meth analyzeImport(Info Xs)
	    proc {CheckFeatures Fs Functor}
	       case Fs
	       of F|Fr then
		  case F
		  of A#V then
		     {self.symTable startScope(name:V
					       info:importref('import':Functor
							      field:A)
					       env:@currentEnvironment)}
		  else
		     skip
		  end
		  {CheckFeatures Fr Functor}
	       [] nil then
		  skip
	       end
	    end
	 in
	    case Xs
	    of X|Xr then 
	       {self.symTable
		startScope(name:X.name info:X env:@currentEnvironment)}
	       {CheckFeatures X.fields X.id}
	       InfoClass, analyzeImport(Info Xr)
	    [] nil then
	       skip
	    end
	 end
	 meth analyzeDefine(Xs)
	    proc {CheckElement X}
	       case X
	       of fLocal(S1 S2) then
		  InfoClass, beginDeclaration
		  {CheckVars S1}
		  InfoClass, endDeclaration
		  InfoClass, analyzeDefine(S2)
		  InfoClass, endScope
	       [] fEq(S1 S2) then
		  InfoClass, beginLeftHand
		  InfoClass, analyzeDefine(S1)
		  InfoClass, endLeftHand
		  InfoClass, analyzeDefine(S2)
	       [] fVar(V) then % unknown????
		  InfoClass, checkVariable({ChangeToAtom fVar(V)})
	       [] fFun(fVar(V) Args Comment) then ID Var in
		  Var = {ChangeToAtom fVar(V)}
		  InfoClass, addInfo(Var function(name:Var id:ID
					      args:{List.map Args ChangeToAtom}
						comment:Comment) ID)
	       [] fProc(fVar(V) Args Comment) then ID Var in
		  Var = {ChangeToAtom fVar(V)}
		  InfoClass, addInfo(Var procedure(name:Var id:ID
					      args:{List.map Args ChangeToAtom}
						 comment:Comment) ID)
	       [] fClass(fVar(V) Ds Ms Comment) then
		  Inherits = {Cell.new nil}
		  Features = {Cell.new nil}
		  Refs ID Var
	       in 
		  Var = {ChangeToAtom fVar(V)}
		  InfoClass, analyzeClassDescriptors(Ds Inherits Features _
						     ?Refs)
		  InfoClass, addInfo(Var 'class'(name:Var
						inherits:{Cell.access Inherits}
						features:{Cell.access Features}
						 id:ID methods:Ms
						 comment:Comment
						 refs:Refs) ?ID)
	       [] fFunctor(V Y Comment) then Functor Var in
		  Var = {ChangeToAtom V}
		  InfoClass, pushFunctor()
		  InfoClass, setComment(Comment)
		  InfoClass, setFunctorName(Var)
		  {FunctorDescriptors Y}
		  InfoClass, analyzeFunctor(Functor)
		  InfoClass, popFunctor()
		  InfoClass, addInfo(Var Functor Functor.id)
	       [] theFunctor then Functor N Y Comment in
		  self.functorCode = fFunctor(N Y Comment)
		  InfoClass, pushFunctor()
		  InfoClass, setComment(Comment)
		  InfoClass, setFunctorName(N)
		  {FunctorDescriptors Y}
		  InfoClass, analyzeFunctor(Functor)
		  Functor.id = {Name.new}
		  InfoClass, popFunctor()
		  self.result = Functor
	       else
		  if {Record.is X} then
		     InfoClass, analyzeDefine({Record.toList X})
		  else
		     skip
		  end
	       end
	    end
	    proc {CheckVars Xs}
	       case Xs
	       of fVar(Var)|Xr then
		  if @inDeclaration \= 0 then
		     {self.symTable startScope(name:{ChangeToAtom fVar(Var)}
					       info:empty
					       env:@currentEnvironment)}
		  end
		  {CheckVars Xr}
	       [] fVar(Var) then
		  if @inDeclaration \= 0 then
		     {self.symTable startScope(name:{ChangeToAtom fVar(Var)}
					       info:empty
					       env:@currentEnvironment)}
		  end
	       [] X|Xr then
		  {CheckElement X}
		  {CheckVars Xr}
	       [] nil then
		  skip
	       else
		  {CheckElement Xs}
	       end
	    end
	 in
	    case Xs
	    of X|Xr then
	       {CheckElement X}
	       InfoClass, analyzeDefine(Xr)
	    [] nil then
	       skip
	    else
	       {CheckElement Xs}
	    end
	 end
	 %% Splits class descriptors in Inherits, Features and Others(Os)
	 meth analyzeClassDescriptors(Xs Inherits Features ?Os ?Refs)
	    % Y == Ref ?????????
	    proc {AnalyzeExpression X ?Y ?Ref}
	       case X
	       of fVar(V) then Info Id in
		  {self.symTable condGetScope(name:{ChangeToAtom fVar(V)}
					      info:Info id:Id
					      default:empty)}
		  if Info == empty then
		     Ref = {Name.new}
		     Y = Ref
		     {Dictionary.put self.infoTable Ref
		      unknown(name:{ChangeToAtom fVar(V)})}
		  else
		     Ref = Id
		     {Dictionary.put self.infoTable Ref Info}
		     case Info
		     of importref(...) then
			if {Not {List.member Ref @importrefs}} then
			   importrefs <- Ref|@importrefs
			end
		     else
			skip
		     end
		     Y = Ref
		  end
	       [] fOpApply('.' [fVar(V) fAtom(A)]) then Info in
		  {self.symTable condGetScope(name:{ChangeToAtom fVar(V)}
					      info:Info default:not_found)}
		  if Info == not_found then
		     Ref = {Name.new}
		     Y = Ref
		     {Dictionary.put self.infoTable Ref
		      unknown(name:{ChangeToAtom X})}
		  else Feature in
		     case Info
		     of 'import'(...) then
			Ref = {Name.new}
			Feature = importref('import':Info.id
					    field:{ChangeToAtom A})
		     [] 'require'(...) then
			Ref = {Name.new}
			Feature = importref('import':Info.id
					    field:{ChangeToAtom A})
		     else
			Ref = {Name.new}
			{Dictionary.put self.infoTable Ref
			 unknown(name:{ChangeToAtom X})}
			Feature = no_feature
		     end
		     if Feature \= no_feature then
			{Dictionary.put self.infoTable Ref Feature}
			if {Not {List.member Ref @importrefs}} then
			   importrefs <- Ref|@importrefs
			end
		     end
		     Y = Ref
		  end
	       else
		  Y = X
		  Ref = no_ref     
	       end
	    end
	    proc {AnalyzeSequence Xs ?Ys ?Refs}
	       case Xs
	       of X|Xr then Y Yr Ref Refs1 in
		  Ys = Y|Yr 
		  {AnalyzeExpression X ?Y ?Ref}
		  if Ref == no_ref then
		     Refs = Refs1
		  else
		     Refs = Ref|Refs1
		  end
		  {AnalyzeSequence Xr ?Yr ?Refs1}
	       [] nil then
		  Ys = nil
		  Refs = nil
	       else Ref in
		  {AnalyzeExpression Xs ?Ys ?Ref}
		  if Ref == no_ref then
		     Refs = nil
		  else
		     Refs = [Ref]
		  end
	       end
	    end
	 in
	    case Xs
	    of X|Xr then
	       case X
	       of fFrom(From) then Refs1 Refs2 I in
		  {AnalyzeSequence From I ?Refs1}
		  {Cell.assign Inherits I}
		  InfoClass, analyzeClassDescriptors(Xr unit Features
						     ?Os ?Refs2)
		  Refs = {Union Refs1 Refs2}
	       [] features(F) then
		  {Cell.assign Features F}
		  InfoClass, analyzeClassDescriptors(Xr Inherits unit
						     ?Os ?Refs)		  
	       else Or in
		  Os = X|Or
		  InfoClass, analyzeClassDescriptors(Xr Inherits Features
						     ?Or ?Refs)
	       end
	    [] nil then
	       Os = nil
	       Refs = nil
	    end
	 end
	 meth exportItems(Xs ?Ys ?Refs)
	    case Xs
	    of fExportItem(A#fVar(V))|Xr then Export Id  Y Yr Refs1 in
	       {self.symTable condGetScope(name:{ChangeToAtom fVar(V)}
					   info:Export id:Id
					   default:empty)}
	       if {Dictionary.member self.infoTable Id} then
		  skip
	       else
		  if Export == empty then
		     {Dictionary.put self.infoTable Id
		      unknown(name:{ChangeToAtom fVar(V)})}
		  else 
		     {Dictionary.put self.infoTable Id Export}
		  end
	       end
	       Ys = Y|Yr
	       Y = 'export'(name:{ChangeToAtom A} ref:Id id:{Name.new})
	       Refs = Id|Refs1
	       InfoClass, exportItems(Xr ?Yr ?Refs1)
	    [] fExportItem(fVar(V))|Xr then Export Id Y Yr Refs1 in
	       {self.symTable condGetScope(name:{ChangeToAtom fVar(V)}
					   info:Export id:Id
					   default:empty)}
	       if {Dictionary.member self.infoTable Id} then
		  skip
	       else 
		  if Export == empty then
		     {Dictionary.put self.infoTable Id
		      unknown(name:{ChangeToAtom fVar(V)})}
		  else
		     {Dictionary.put self.infoTable Id Export}
		  end
	       end
	       Ys = Y|Yr
	       Y = 'export'(name:unit ref:Id id:{Name.new})
	       Refs = Id|Refs1
	       InfoClass, exportItems(Xr ?Yr ?Refs1)
	    [] nil then
	       Ys = nil
	       Refs = nil
	    end
	 end
	 meth beginDeclaration
	    inDeclaration <- @inDeclaration + 1
	    environmentStack <- @currentEnvironment|@environmentStack
	    currentEnvironment <- {Name.new}
	 end
	 meth endDeclaration
	    inDeclaration <- @inDeclaration - 1
	 end
	 meth endScope
	    Stack Env
	 in
	    {self.symTable endScope(env:@currentEnvironment)}
	    Env|Stack = @environmentStack
	    currentEnvironment <- Env
	    environmentStack <- Stack
	 end
	 meth beginLeftHand
	    if @inDeclaration \= 0 then
	       leftHand <- true
	    end
	 end
	 meth endLeftHand
	    if @inDeclaration \= 0 then
	       leftHand <- false
	    end
	 end
	 meth checkVariable(Var)
	    if @inDeclaration \= 0 andthen @leftHand then
	       {self.symTable startScope(name:Var info:empty
				       env:@currentEnvironment)}
	    end
	 end
	 meth addInfo(Var Info ID <= _)
	    if @inDeclaration \= 0 then
	       {self.symTable startScope(name:Var info:Info id:ID
					 env:@currentEnvironment)}
	    else
	       try
		  {self.symTable updateScope(name:Var info:Info id:ID)}
	       catch noScopeFor(Var) then % first appearance
		  {self.symTable startScope(name:Var info:Info id:ID
					    env:@currentEnvironment)}
	       end
	    end
	    if @inPrepare andthen {Not {List.member ID @refsInPrepare}} then
	       refsInPrepare <- ID|@refsInPrepare
	    end
	 end
      end

      fun {FindFunctor Xs Found}
	 case Xs
	 of X|_ then  % anything comming after the functor is not of interest
	    case X
	    of fFunctor(...) = FC then
	       {Info setFunctorCode(FC)}
	       {Cell.assign Found true}
	       theFunctor
	    [] fLocal(S1 S2) then
	       fLocal(S1 {FindFunctor S2 Found})
	    else
	       unit
	    end
	 [] fFunctor(...) = FC then
	    {Info setFunctorCode(FC)}
	    {Cell.assign Found true}
	    theFunctor
	 [] nil then
	    nil
	 else
	    unit
	 end
      end
   in
      proc {Initialize}
	 {Cell.assign InfoObject {New InfoClass init(returnAll:false)}}
      end
      proc {Analyze C1 Out} Found = {Cell.new false} C2 in
	 C2 = {FindFunctor C1 Found}
	 if {Cell.access Found} then
	    {Info analyze(C2 Out)}
	 else
	    raise analyze_error('No functor is found') end
	 end
      end	 
   end
   
   fun {ParseFile FileName Reporter GetSwitch Macros}
      \gumpscannerprefix analyzeFunctor
      MyScanner = {New OzScanner init(gump:{GetSwitch gump}
				      showInsert:{GetSwitch showinsert}
				      reporter:Reporter macros:Macros)}
      MyParser = {New OzParser init(allowDeprecated:{GetSwitch allowdeprecated}
				    'scanner':MyScanner
				    reporter:Reporter)}
      Status ParseTree
   in
      {Initialize}
      {MyScanner scanFile(FileName)}
      {MyParser parse(file(?ParseTree) ?Status)}
      {MyScanner close()}
      if Status then
	 {Analyze ParseTree}
      else
	 raise parse_error end
      end
   end

   fun {ParseVirtualString VS Reporter GetSwitch Macros}
      \gumpscannerprefix analyzeFunctor
      MyScanner = {New OzScanner init(gump:{GetSwitch gump}
				      showInsert:{GetSwitch showinsert}
				      reporter:Reporter macros:Macros)}
      MyParser = {New OzParser init(allowDeprecated:{GetSwitch allowdeprecated}
				    'scanner':MyScanner
				    reporter:Reporter)}
      Status ParseTree
   in
      {Initialize}
      {MyScanner scanVirtualString(VS)}
      {MyParser parse(file(ParseTree) ?Status)}
      {MyScanner close()}
      if Status then
	 {Analyze ParseTree}
      else
	 raise parse_error end
      end
   end
end
