%%%
%%% Authors:
%%%   Nils Franzen (nils.franzen@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzen, 1999
%%%   Andreas Sundstrom, 1999
%%%
%%% Last change:
%%%   $Date: 2000/02/22 10:35:03 $ by $Author: nilsf $
%%%   $Revision: 1.12 $
%%%
%%% See the file "LICENSE" or
%%%    http://mozart.ps.uni-sb.de/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%
/**
% Traverse the inheritance tree for all classes and write an inheritance
% graph toghether with an inheritance free version of the class
*/

functor
import
   Browser(browse:Browse)
   System
   HTML
   Comments(analyzeComment:AnalyzeComment
	    mc:MC)
export
   AnalyzeClasses
define
   A2S=Atom.toString
   VSS=VirtualString.toString

   %% Make records '#'(...) of lists [...]
   fun{MMap Ls F}
      {List.toRecord '#' {List.mapInd {Map Ls F} fun{$ I X} I#X end}}
   end

   ClassDB={NewDictionary}
   
   fun{TraverseInheritance C Is X0 Y0 X1 Y1 BX}
      L={Max {Length {A2S C.name}}+1+X0 BX}
   in
      if {Label C}\='class' orelse C.inherits==nil then
	 X1=L
	 Y1=Y0
	 {Adjoin coords(x:X0#X1 y:Y0#Y1) C}
      else
	 LastX={NewCell X0}
	 NY=Y0+1
	 MaxY={NewCell NY}
	 C2
	 Single=({Length C.inherits}==1)
      in
	 C2={Map C.inherits
	     fun{$ Key}
		X=try Is.Key catch _ then {Browse error(Key Is)} {Wait _} nil end
	     in
		case {Label X}
		of 'class' then
		   X2 Y2
		   A={TraverseInheritance X {Dictionary.get ClassDB X.id} {Access LastX} NY X2 Y2 if Single then L else 0 end}
		in
		   {Assign LastX X2}
		   {Assign MaxY {Max {Access MaxY} Y2}}
		   A
		[] unknown then OX NX in
		   {Exchange LastX OX NX}
		   NX={Length {A2S X.name}}+1+OX
		   unknown(name:X.name x:OX#NX y:NY#NY)
		end
	     end}
	 X1={Max {Access LastX} L}  Y1={Access MaxY}
	 {Adjoin coords(x:X0#X1 y:Y0#Y1) {AdjoinAt C inherits C2}}
      end
   end

   local
      %% Top-left corner is 0,0
      class CharArray
	 feat arrayx x y meta
	 meth init(x:X y:Y)
	    self.x=X self.y=Y
	    self.arrayx={NewArray 0 X nil}
	    self.meta={NewDictionary}
	    {For 0 X 1 proc{$ I} {Array.put self.arrayx I {NewArray 0 Y nil}} end}
	 end
      
	 meth put(V x:X y:Y)
	    if {IsChar V} then {Array.put {Array.get self.arrayx X} Y V}
	    elseif {IsRecord V} andthen {Label V}==meta then {Array.put {Array.get self.arrayx X} Y V}
	    else raise notChar(V) end end
	 end
	 meth get(V x:X y:Y) V={Array.get {Array.get self.arrayx X} Y} end
	 meth insertStr(S x:X y:Y meta:Meta<=unit)
	    {List.forAllInd {VirtualString.toString S} proc{$ I C} {self put(C x:X+I-1 y:Y)} end}
	    if Meta\=unit then Key={VirtualString.toAtom X#":"#Y} in
	       {Dictionary.put self.meta Key Meta}
	    end
	 end
	 meth getRow(R y:Y meta:M<=false)
	    R=if M==false then Ls={List.make self.x+1} in
		 {List.forAllInd Ls proc{$ I X} {self get(X x:I-1 y:Y)} end}
		 {Map {Reverse {List.dropWhile {Reverse Ls} fun{$ C} C==nil end}} fun{$ X} if X==nil then &  else X end end}
	      else
		 Ignore={NewCell 0} Tmp={NewName}
		 Ls Ls1={List.make self.x+1}
	      in
		 {List.forAllInd Ls1 proc{$ I X} O N Key={VirtualString.toAtom (I-1)#":"#Y} in
					{Exchange Ignore O N}
					if O>0 then
					   N=O-1
					   X=Tmp
					else
					   N=O
					   if {Dictionary.member self.meta Key} then
					      GetMeta={Dictionary.get self.meta Key}
					      Skip Str={GetMeta Skip}
					   in
					      {Assign Ignore Skip}
					      X=Str
					   else
					      {self get(X x:I-1 y:Y)}
					   end
					end
				     end}
		 Ls={Filter Ls1 fun{$ X} X\=Tmp end}
		 {MMap {Reverse {List.dropWhile {Reverse Ls} fun{$ C} C==nil end}} fun{$ X}
										      if X==nil then " " 
										      elseif {IsRecord X} then X
										      elseif X==&  then " "
										      else {Char.toAtom X}
										      end
										   end}
	      end
	 end
	 meth dump(D meta:M<=false) Ls={List.make self.y+1} in
	    {List.forAllInd Ls proc{$ I X} {self getRow(X y:I-1 meta:M)} end}
	    D={Reverse Ls}
	 end
      end
   
      fun{GetXY Cs What}
	 L={Length {A2S Cs.name}}-1
	 DX=Cs.x.2-Cs.x.1
	 TV=((DX-L)+1) div 2
	 Y=(Cs.y.1)*4
	 X=(Cs.x.1)+TV
      in
	 case What of parent then (X+(L div 2))#(Y+1)
	 [] child then(X+(L div 2))#(Y-1)
	 [] normal then X#Y
	 end
      end
   
      proc{Insert Ar What X Y}
	 O={Ar get($ x:X y:Y)}
      in
	 if O\=&+ then
	    if (O==&` andthen What==&') orelse (O==&' andthen What==&`) then
	       {Ar insertStr("+" x:X y:Y)}
	    elseif (O==&` orelse O==&') andthen What==&- then
	       skip
	    else
	       {Ar insertStr([What] x:X y:Y)}
	    end
	 end
      end

      proc{DrawConnection Ar A B}
	 A1={GetXY A parent}
	 B1={GetXY B child}
	 BX=B1.1
	 BY=B1.2
	 AX=A1.1
	 AY=A1.2 
	 DX=AX-BX
	 MY=BY-1 %% Y between A and B
      in
	 if AX<0 orelse BX<0 then
	    raise negXX(AX BX DX) end
	 end
	 if AY<0 orelse BY<0 then
	    raise negYY(AY BY DX) end
	 end
	 {Insert Ar &| AX AY}
	 {Insert Ar &| BX BY}
	 case DX of 0 then
	    {Insert Ar &| BX MY}
	 else
	    MiX={Min AX BX}
	    MaX={Max AX BX}
	 in
	    if DX<0 then {Insert Ar &' BX MY}
	    else {Insert Ar &` BX MY}
	    end
	    {Insert Ar &+ AX MY}
	    {For MiX+1 MaX-1 1 proc{$ X} {Insert Ar &- X MY} end}
	 end
      end
   
      proc{Traverse Cs Ar}
	 X#Y={GetXY Cs normal}
	 Name={A2S Cs.name}
      in
	 {Ar insertStr(Name x:X y:Y
		       meta:proc{$ Skip MetaStr}
			       Skip={Length Name}-1
			       MetaStr=if {HasFeature Cs href} then
					  a('class':"pre" href:Cs.href Name)
				       else
					  Name
				       end
			    end)}
	 if {HasFeature Cs inherits} then
	    {ForAll Cs.inherits proc{$ X}
				   {Traverse X Ar}
				   {DrawConnection Ar Cs X}
				end}
	 end
      end
      
      fun{BeginDisplay Cs X Y}
	 Ar={New CharArray init(x:X y:Y*4)}
      in
	 {Traverse Cs Ar}
	 Ar
      end
   in
      fun{ClassHierarchyHtml C} Ar={BeginDisplay C C.x.2 C.y.2} in
	 {Ar dump($ meta:true)}
      end

%       fun{Ascii C} Ar={BeginDisplay C C.x.2 C.y.2} in
% 	 {Ar dump($)}
%       end      
   end

   proc{DoHTML File C CHs CMs Is OutDir Stylesheet}
      A={MakeHTML C CHs CMs Is Stylesheet}
   in
      {System.showInfo "Writing "#OutDir#File}
      {HTML.out OutDir#File A}
   end

   proc{MakeHTML C CH CM IndexL Stylesheet Body} B in
      B=body("\n"
	     p('class':"margin"
	       table(border:"0" cellspacing:"0" cellpadding:"0" "\n"
		     tr(td(width:"100" hr(width:"100"))) "\n"
		     {Record.map IndexL fun{$ X} '#'(tr(td(width:"100" X)) "\n") end}
		     tr(td(width:"100" hr(width:"100"))) "\n"
		    ))
	     "\n"
	     pre("\n" {MMap CH fun{$ X} X#"\n" end})
	     "\n"
	     CM
	     "\n"
	    )
      Body=html("\n "
		head("\n  " title({VSS "Class "#C.name}) "\n  "
		     link(href:Stylesheet rel:"stylesheet" type:"text/css")
		     "\n ")
		"\n" B
		"\n")
   end

   fun{ProcessClasses CsD} Cs#DB=CsD in
      %% Insert all classes into ClassDB, so we later can get their resolved refs
      {ForAll Cs proc{$ X} NodeID#C=X in
		    {Dictionary.put ClassDB C.id {List.toRecord resolved {Dictionary.condGet DB NodeID nil}}}
		 end}
      
      %% Process all classes
      {Map {Reverse Cs} fun{$ X} NodeID#C=X in
			   NodeID#{TraverseInheritance C {Dictionary.get ClassDB C.id} 0 0 _ _ 0}
			end}
   end

   proc{GetAllMethsAndFeat C MDB FDB}
      if {Label C}=='class' then
	 {ForAll C.inherits proc{$ X} {GetAllMethsAndFeat X MDB FDB} end}
	 {ForAll C.features proc{$ X} {Dictionary.put FDB X X} end}
	 {ForAll C.methods proc{$ M} {Dictionary.put MDB M.name M} end}
      end
   end
   
   fun{CollectMethods C}
      MDB={NewDictionary}
      FDB={NewDictionary}
      TypeDB={NewDictionary}

      fun{DisplayArg I N} 
	 case N of A1#Arg then
	    A=if {IsInt A1} then
		 {VirtualString.toAtom A1#nil}
	      elseif {IsName A1} then
		 {VirtualString.toAtom "<N:"#I#">"}
	      else A1 end
	 in
	    {VSS
	     %% Feat...
	     if I==A1 then nil else A#":" end#
	     %%  followed by value
	     {Dictionary.condGet TypeDB A Arg}}
	 end
      end
      fun{GetArgs M S Cm}
	 Index={NewCell 1}
      in
	 {Dictionary.removeAll TypeDB}
	 {ForAll Cm.argtypes proc{$ X} {Dictionary.put TypeDB X.1 X.2} end}
	 {ForAll Cm.extargtypes proc{$ X} {Dictionary.put TypeDB X.2 X.3} end}
	 {List.foldL S fun{$ O N} OI NI in
			  {Exchange Index OI NI}
			  NI=OI+1
			  if O==nil then {DisplayArg OI N}
			  else O#" "#{DisplayArg OI N} end
		       end nil}#
	 if M==open then
	    if S\=nil then "&nbsp;..." else "..." end
	 else nil end
      end
   in
      {GetAllMethsAndFeat C MDB FDB}
      code("\n"
	   %% Comments for the class
	   {MC {AnalyzeComment C.comment}}

	   %% Display class name and inheritance
	   span('class':"keyword" "class") " " span('class':"type" {A2S C.name})  " (Inheritance-free)" br "\n"

	   %% Show features
	   if {Dictionary.items FDB}\=nil then
	      '#'("&nbsp;&nbsp;&nbsp;" span('class':"keyword" "feat")
		  {MMap {Sort {Dictionary.items FDB} fun{$ N O} N<O end} fun{$ F} {VSS "&nbsp;"#F} end}
		  "\n")
	   else nil end
	   br "\n"
	   ul({MMap
	       {Sort {Dictionary.items MDB} fun{$ N O} N.name<O.name end}
	       fun{$ X}
		  Cm={AnalyzeComment X.comment}
		  Args={Record.toListInd X.args}
	       in
		  li('#'(span('class':"functionname" {A2S X.name})
			 "(" {GetArgs {Label X.args} Args Cm} ")" br "\n"
			 {MC Cm}))
	       end})
	   "\n"
	   span('class':"keyword" "end")
	   "\n"
	  )
   end
   
   proc{AnalyzeClasses Is OutDir Stylesheet Cs}
      {ForAll {ProcessClasses Cs}
       proc{$ C}
	     {DoHTML "class"#C.1#".html" C.2 {ClassHierarchyHtml C.2} {CollectMethods C.2} Is OutDir Stylesheet}
       end}
   end
end
