%%%
%%% Authors:
%%%   Nils Franzen (nils.franzen@sics.se)
%%%   Simon Lindblom (simon@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzen, 1999
%%%   Andreas Sundstrom, 1999
%%%   Simon Lindblom, 1999
%%%
%%% Last change:
%%%   $Date: 1999/12/09 10:36:21 $ by $Author: nilsf $
%%%   $Revision: 1.5 $
%%%
%%% See the file "LICENSE" or
%%%    http://mozart.ps.uni-sb.de/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
% Extract arguments from comments
*/

functor
import
   Regex at 'x-oz://contrib/regex'
export
   AnalyzeComment
   FixComments
   AnalyzeCommentCVS
   DisplayExtArgs
   DisplayInfo
   mc:MC
define
   VSS=VirtualString.toString
   %% Extracting info from comments

   fun{MC Cm}
      '#'("\n" {DisplayInfo Cm.info Cm.cvs} "\n" span('class':"string" {FixComments Cm.string})
	  "\n" {DisplayExtArgs Cm.extargtypes}) 
   end
   

   local
      WS="[ \t]*"
      Comment_RE={Regex.make "/[*]{2}((\n|.)+)[*]/"}
      fun{GetComment BS} C={Regex.search Comment_RE BS} in
	 if C==false then nil
	 else {ByteString.slice BS (C.0.1+3) (C.0.2-2)} end
      end
      Arg_RE={Regex.make "@([a-zA-Z0-9_-]+)"#WS#"="#WS#"([a-zA-Z0-9/?_]+)"}
      fun{GetArguments BS}
	 Cs={Regex.allMatches Arg_RE BS}
      in
	 {Map Cs fun{$ C}
		    A={Regex.group 1 C BS}
		    V={Regex.group 2 C BS}
		 in
		    {String.toAtom {ByteString.toString A}}#{ByteString.toString V}
		 end}
      end   
      Ext_RE={Regex.make "@(arg|ret)"#WS#"([a-zA-Z0-9_-]+)"#WS#"([a-zA-Z0-9_-]+)"#WS#"(.+)\n"}
      fun{GetExtArguments BS NewST}
	 EmptyBS={ByteString.make ""}
	 Cs={Regex.allMatches Ext_RE BS}
	 NewBS Ind
      in
	 if Cs==nil then Ind=0 NewBS=EmptyBS
	 else
	    Ind#NewBS={List.foldL Cs fun{$ A C}
					Ind#T=A 
					Tmp={ByteString.append T {ByteString.slice BS Ind (C.0.1)}}
				     in
					(C.0.2)#Tmp
				     end 0#EmptyBS}
	 end
	 if BS==nil then NewST=EmptyBS else
	    NewST={ByteString.append NewBS {ByteString.slice BS Ind {ByteString.length BS}}}
	 end
	 {Map Cs fun{$ C}
		    A={Regex.group 1 C BS}
		    I={Regex.group 2 C BS}
		    N={Regex.group 3 C BS}
		    T={Regex.group 4 C BS}
		 in
		    {String.toAtom {ByteString.toString A}}#
		    {String.toAtom {ByteString.toString I}}#
		    {String.toAtom {ByteString.toString N}}#
		    {ByteString.toString T}
		 end}
      end
      
      Info_RE={Regex.make "@(version|author|date)"#"(.+)\n"}
      fun{GetInfo BS NewST}
	 EmptyBS={ByteString.make ""}
	 Cs={Regex.allMatches Info_RE BS}
	 NewBS Ind
      in
	 if Cs==nil then Ind=0 NewBS=EmptyBS
	 else
	    Ind#NewBS={List.foldL Cs fun{$ A C}
					Ind#T=A 
					Tmp={ByteString.append T {ByteString.slice BS Ind (C.0.1)}}
				     in
					(C.0.2)#Tmp
				     end 0#EmptyBS}
	 end
	 if BS==nil then NewST="" else
	    NewST={ByteString.toString {ByteString.append NewBS{ ByteString.slice BS Ind {ByteString.length BS}}}}
	 end
	 {List.toRecord info {Map Cs fun{$ C}
					A={Regex.group 1 C BS}
					T={Regex.group 2 C BS}
				     in
					{String.toAtom {ByteString.toString A}}#
					{ByteString.toString T}
				     end}}
      end
      Cvs_RE={Regex.make "\\$(Date:|Author:|Revision:)"#WS#
	      "([a-zA-Z0-9_\\:\\/\\. ]+)"#WS#"\\$"}
      fun{GetCVS BS}
	 Cs={Regex.allMatches Cvs_RE BS}
      in
	 {List.toRecord cvs {Map Cs fun{$ C}
				       A={Regex.group 1 C BS}
				       T={Regex.group 2 C BS}
				    in
				       {String.toAtom {ByteString.toString A}}#
				       {ByteString.toString T}
				    end}}
      end
	 
      fun{DoAnalyzeComment Txt ParseCVS} 
	 BS={ByteString.make Txt}
	 CommentBeforeParsing1={GetComment BS}
	 CommentBeforeParsing2
	 Comment
	 ExtArgs={GetExtArguments CommentBeforeParsing1 CommentBeforeParsing2}
	 Info={GetInfo CommentBeforeParsing2 Comment}
	 CVS
      in
	 if ParseCVS then CVS={GetCVS BS} 
	 else CVS=nil end
	 comment(string:Comment extargtypes:ExtArgs cvs:CVS
		 info:Info argtypes:{GetArguments BS})
      end
   in
      fun{AnalyzeComment Txt} {DoAnalyzeComment Txt false} end
      fun{AnalyzeCommentCVS Txt CVS} {DoAnalyzeComment Txt CVS} end
   end
   %% Insert an <BR> at each newline
   fun{FixComments S}
      Ss={String.tokens {Reverse {List.dropWhile {Reverse {List.dropWhile S Char.isSpace}} Char.isSpace}} &\n}
        %% Remove (WS)%+ in front of each row
      fun{RemovePercent X}
	 WS="[ \t]*"
	 EmptyBS={ByteString.make ""}
	 NewST
	 Arg_RE={Regex.make "^"#WS#"%+"}
	 fun{GetPercents BS}
	    Cs={Regex.allMatches Arg_RE BS}
	    NewBS Ind
	 in
	    if Cs==nil then NewST={ByteString.toString BS} else
	       Ind#NewBS={List.foldL Cs fun{$ A C}
					   Ind#T=A 
					   Tmp={ByteString.append T
						{ByteString.slice BS Ind (C.0.1)}}
					in
					   (C.0.2)#Tmp
					end 0#EmptyBS}
	       if BS==nil then NewST="" else
		  NewST={ByteString.toString
			 {ByteString.append NewBS
			  {ByteString.slice BS Ind {ByteString.length BS}}}}
	       end
	    end
	    NewST
	 end
      in
	 {GetPercents {ByteString.make X}}
      end
   in
      {VSS {FoldR Ss fun{$ N O}	{RemovePercent N}#"<BR>\n"#O end nil}}
   end

   %% Display extended argumendescription
   fun{DisplayExtArgs A}
      if A==nil then "" else
	 '#'({List.toRecord table {Append [border#0 cellspacing#2 cellpadding#2]
				   {List.mapInd {Append [tr(th(colspan:4 bgcolor:"#cccc99" "Arguments"))]
						 {List.foldR A
						  fun{$ X Acc} Q#W#E#R=X in
						     tr(bgcolor:"#eeeecc"
							th(bgcolor:"#cccc99" {Atom.toString Q} )
							td({Atom.toString W})
							td({Atom.toString E})
							td( R))
						     |Acc
						  end nil}} fun{$ Ind X} Ind#X end}}} "\n")
      end
   end

   %% Display info tags
   fun{DisplayInfo A B}
      table(border:0 cellspacing:2 cellpadding:2
	    tr(bgcolor:"#eeeecc" 
	       if {HasFeature A version} then
		  td({VirtualString.toString "Version: "#A.version#"\n"})
	       else "\n" end
	       if {HasFeature A date} then
		  td({VirtualString.toString "Date: "#A.date#"\n"})
	       else "\n" end
	       if {HasFeature A author} then
		  td({VirtualString.toString "Author: "#A.author#"\n"})
	       else "\n" end
	       if {HasFeature B 'Revision:'} then
		  td({VirtualString.toString "Revision: "#B.'Revision:'#
		      "\n"})
	       else "\n" end      
	       if {HasFeature B 'Date:'} orelse {HasFeature B 'Author:'}
	       then Tmp=if { HasFeature B 'Date:'} then B.'Date:'
			else "" end#if { HasFeature B 'Author:'}
				    then " by "#B.'Author:' else "" end
	       in
		  td({VirtualString.toString "Last Modified: "#Tmp#
		      "\n"})
	       else "\n" end      
	      ))
   end
end
