%%%
%%% Authors:
%%%   Nils Franzen (nils.franzen@sics.se)
%%%   Andreas Sundstrom (andreas@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzen, 1999
%%%   Andreas Sundstrom, 1999
%%%
%%% Last change:
%%%   $Date: 1999/12/09 10:36:22 $ by $Author: nilsf $
%%%   $Revision: 1.4 $
%%%
%%% See the file "LICENSE" or
%%%    http://mozart.ps.uni-sb.de/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
% Takes a simplified description of all analyzed functors (name, imports and exports).
% Creates a new window and display the relation between the functors.
% Lets the user arrange the functors and save the final picture as a postscrip file.
*/

functor
import
   Tk
   System(show:Show)
   Browser(browse:Browse)
export
   Display
define
   PI=3.141
   PIPI=PI*2.0
   local
      FTI=Float.toInt
      ITF=Int.toFloat
      fun{CCW P0 P1 P2}
	 Dx1 = P1.1 - P0.1
	 Dx2 = P2.1 - P0.1
	 Dy1 = P1.2 - P0.2
	 Dy2 = P2.2 - P0.2
      in
	 if     Dx1*Dy2 > Dy1*Dx2 then 1
	 elseif Dx1*Dy2 < Dy1*Dx2 then ~1
	 elseif (Dx1*Dx2 < 0) orelse (Dy1*Dy2 < 0) then ~1 
	 elseif ({Pow Dx1 2} + {Pow Dx2 2}) < ({Pow Dy1 2} + {Pow Dy2 2}) then 1
	 else 0 end
      end
   in
      fun{Dist P1 P2}
	 {FTI {Sqrt {ITF {Pow (P1.1 - P2.1) 2} + {Pow (P1.2 - P2.2) 2}}}}
      end
      fun{Intersect L1 L2}
	 if {CCW L1.1 L1.2 L2.1} * {CCW L1.1 L1.2 L2.2} < 0 andthen 
	    {CCW L2.1 L2.2 L1.1} * {CCW L2.1 L2.2 L1.2} < 0 then
	    true
	 else false end
      end
   end

   proc{SavePostscript DC}
      case {Tk.return tk_getSaveFile(title:"Saving postscript..."
				     parent:DC
				     initialfile:"functormap.ps"
				     filetypes:q(q('Ps' q('.ps'))
						 q('All Files' '*')
						)
				    )}
      of nil then skip
      elseof File then [X0 Y0 X1 Y1]={DC tkReturnListInt(bbox alltag $)} in
	 {DC tk(postscript file:File x:X0 y:Y0 width:X1-X0 height:Y1-Y0 colormode:color)}
      end
   end
   
   proc{Display DFs}
      DONE
      T={New Tk.toplevel tkInit(title:"Functor Map" delete:proc{$} {T tkClose} DONE=unit end)}
      DC={New Tk.canvas tkInit(parent:T bg:white width:500 height:600)}
      SX={New Tk.scrollbar tkInit(parent:T orient:horizontal width:8)}
      SY={New Tk.scrollbar tkInit(parent:T orient:vertical width:8)}
      {Tk.addXScrollbar DC SX}
      {Tk.addYScrollbar DC SY}
      BF={New Tk.frame tkInit(parent:T)}
      B1={New Tk.button tkInit(parent:BF text:"Quit" action:proc{$} {T tkClose} DONE=unit end)}
      B2={New Tk.button tkInit(parent:BF text:"Save Postscript" action:SavePostscript#DC)}
      {Tk.batch [grid(DC row:0 column:0 sticky:news)
		 grid(SX row:1 column:0 sticky:we)
		 grid(SY row:0 column:1 sticky:ns)
		 grid(BF row:2 column:0 columnspan:2 sticky:we)
		 grid(B1 row:0 column:0 sticky:we)
		 grid(B2 row:0 column:1 sticky:we)
		 grid(columnconfigure T 0 weight:1)
		 grid(rowconfigure T 0 weight:1)
		 grid(columnconfigure BF 0 weight:1)
		 grid(columnconfigure BF 1 weight:1)]}
      proc{SetScrollRegion} [X0 Y0 X1 Y1]={DC tkReturnListInt(bbox alltag $)} in
	 {DC tk(configure scrollregion:q(X0-5 Y0-5 X1+5 Y1+5))}
      end
      class Box
	 feat tag f imps 
	 attr x y dx dy imp impby:nil angle:0.0 imp2
	 meth init(F text:T x:X y:Y) T1={New Tk.canvasTag tkInit(parent:DC)} 
	    %% Create functor title
	    {DC tk(crea text X Y text:T fill:black tags:T1)}
	    %% Box containing then load
	    [X0 Y0 X1 Y1]={DC tkReturnListInt(bbox T1 $)}
	    T2={New Tk.canvasTag tkInit(parent:DC)}
	    DX=4 DY=3 S=3
	 in
	    %% Shadow
	    {DC tk(crea rect X0-DX+S Y0-DY+S X1+DX+S Y1+DY+S
		   width:0
		   outline:nil
		   fill:grey
		   stipple:gray50
		   tags:T2)}
	    %% Box
	    {DC tk(crea rect X0-DX Y0-DY X1+DX Y1+DY
		   fill:wheat
		   outline:black
		   width:1
		   tags:q(alltag T2))}
	 
	    {DC tk(lower T2 T1)}
	    {DC tk(addtag T1 withtag T2)}
	 
	    x<-X y<-Y
	    self.f=F
	    self.tag=T1
	    imp<-{Filter {Map {Append F.'import' F.'require'} fun{$ X} {String.toAtom X.filename} end} fun{$ X} X\='' end} 
	    self.imps={Length @imp}
	    {T1 tkBind(event:'<3>'
		       action:proc{$}
				 L=self.f.name
			      in
				 {Browse L(x:@x y:@y angle:@angle imp:@imp impby:@impby)}
			      end)}

	    %% Bind movement of nodes
	    {T1 tkBind(event:'<Button-1>'
		       args:[int(x) int(y)]
		       action:proc{$ CX CY}
				 X={DC tkReturnInt(canvasx CX $)}
				 Y={DC tkReturnInt(canvasy CY $)}
			      in
				 @dx=@x-X
				 @dy=@y-Y
			      end)}
      
	    {T1 tkBind(event:'<B1-Motion>'
		       args: [int(x) int(y)]
		       action:proc{$ CX CY}
				 X={DC tkReturnInt(canvasx CX $)}
				 Y={DC tkReturnInt(canvasy CY $)}
				 OX=@x
				 OY=@y
			      in
				 x<-OX+(X-OX+@dx)
				 y<-OY+(Y-OY+@dy)
				 {self.tag tk(move @x-OX @y-OY)}
			      end)}
      
	    {T1 tkBind(event:'<ButtonRelease-1>'
		       action:proc{$}
				 dx<-_
				 dy<-_
				 {DC tk(delete linetag)}
				 {ForAll Os proc{$ N} {N drawConnections} end}
			      end)}
	 end
	 meth Move(dx:DX dy:DY) {self.tag tk(move DX DY)} end
	 meth moveTo(x:X y:Y)
	    {self Move(dx:X-@x dy:Y-@y)}
	    x<-X y<-Y
	 end
	 meth setImportedBy O N in
	    O=imp<-N
	    N={Map O fun{$ X} {Dictionary.get NodeDB X} end}
	    {ForAll N proc{$ X} {X importedBy(self)} end}
	 end
	 meth importedBy(S) O N in
	    O=impby<-N
	    N=S|O
	 end
	 meth placeNearImpBy(C1)
	    Imp=if self.imps==1 andthen {self countImpBy($)}==1 then
		   [@impby.1 @imp.1]
		else
		   @impby
		end
	 in
	    case Imp of [N] then
	       C=C1+100.0
	       A={N getAngle($)}
	       Ay={Float.toInt {Sin A}*C}
	       Ax={Float.toInt {Cos A}*C}
	    in
	       {self setAngle(A)}
	       {self moveTo(x:Ax y:Ay)}
	    elseof [N1 N2] then
	       C=C1+150.0
	       A1={N1 getAngle($)}
	       A2={N2 getAngle($)}
	       Ma={Max A1 A2}
	       Mi={Min A1 A2}
	       Da=Mi+((Ma-Mi) / 2.0)
	       A=if Ma-Mi>PI then
		    Da-PI
		 else
		    Da
		 end
	       Ay={Float.toInt {Sin A}*C}
	       Ax={Float.toInt {Cos A}*C}
	    in
	       {self setAngle(A)}
	       {self moveTo(x:Ax y:Ay)}
	    end
	 end
	 meth countImpBy($) {Length @impby}  end
	 meth getXY($) @x#@y end
	 meth drawConnections MXYC={self getXY($)} in
	    {ForAll @imp proc{$ N}
			    XY1={N getCorner(MXYC $ type:'in')}
			    MXY={self getCorner(XY1 $ type:out)}
			    XY={N getCorner(MXY $ type:'in')}
			 in
			    {DC tk(crea line MXY.1 MXY.2 XY.1 XY.2 fill:red width:1 arrow:last tags:linetag)}
			 end}
	 end
	 meth setAngle(A) angle<-A end
	 meth getAngle($) @angle end
	 meth isBetterPos(XY B)
	    Ds={Map @imp2 fun{$ X} {X getXY($)} end}
	    P={self getXY($)}
	    S1={FoldR Ds fun{$ N O} {Dist P N}+O end 0}
	    S2={FoldR {Map Ds fun{$ P0} if P0==XY then P else P0 end end} fun{$ N O} {Dist XY N}+O end 0}
	 in
	    B=if S2<S1 then true else false end
	 end
	 meth getCorner(XY C type:T<=out)
	    [DX DY]=if T==out then [0 0] else [2 2] end
	    [X0 Y0 X1 Y1]={DC tkReturnListInt(bbox self.tag $)}
	    Mx=X0+((X1-X0) div 2)
	    My=Y0+((Y1-Y0) div 2)
	    MMx=(X1-X0) div 4
	    AllCs=[X0-DX#Y0-DY Mx-MMx#Y0-DY Mx#Y0-DY Mx+MMx#Y0-DY X1+DX#Y1-DY
		   X0-DX#My X1+DX#My
		   X0-DX#Y1+DY Mx-MMx#Y1+DY Mx#Y1+DY Mx+MMx#Y1+DY X1+DX#Y1+DY]
	    C1={FoldR AllCs.2 fun{$ N O} D={Dist XY N} in
				 if D<O.dist then d(N dist:D) else O end
			      end d(AllCs.1 dist:{Dist XY AllCs.1})}
	 in C=C1.1 end
	 meth getLinesCoords($) P={self getXY($)} in {Map @imp2 fun{$ X} P#{X getXY($)} end} end
	 meth dontCountWith(Os)
	    imp2<-{Filter @imp fun{$ X} {Member X Os}==false end}
	 end
      end
      Os Os1 Os2 NodeDB={NewDictionary}
   in
      Os={Map DFs proc{$ F A} 
		     A={New Box init(F text:F.name x:0 y:0)}
		     {Dictionary.put NodeDB F.file A}
		  end}
   
      {ForAll Os proc{$ X} {X setImportedBy} end}
   
      %% Put all "leafs" in Os1 and all nodes that belongs to the circle in Os2
      {List.partition Os fun{$ X}
			    (X.imps==0 andthen {X countImpBy($)}=<2)
			    orelse
			    (X.imps==1 andthen {X countImpBy($)}==1) 
			 end Os1 Os2}

      %% Remove "leafs" (Os1) temporary in the nodes
      {ForAll Os proc{$ X} {X dontCountWith(Os1)} end}

      local
	 fun{CountCrossingLinesFor A As} C={NewCell 0} in
	    {ForAll A proc{$ P1}
			 {ForAll As proc{$ Xs}
				       {ForAll Xs proc{$ P2}
						     if {Intersect P1 P2} then O N in
							{Exchange C O N}
							N=O+1
						     end
						  end}
				    end}
		      end}
	    {Access C}
	 end

	 fun{CountCrossingLines Ns}
	    Ns1={Map Ns fun{$ N} {N getLinesCoords($)} end}
	    OL={NewCell 0}
	 in
	    {List.forAllTail Ns1 proc{$ Ts}
				    case Ts of A|As then
				       O N C={CountCrossingLinesFor A As}
				    in
				       {Exchange OL O N}
				       N=O+C
				    else
				       skip
				    end
				 end}
	    {Access OL}
	 end
	 
	 Repeat={NewCell true}
	 fun{BenefitFromSwap A B}
	    if {A isBetterPos({B getXY($)} $)} andthen {B isBetterPos({A getXY($)} $)} then
	       true
	    else
	       false
	    end
	 end
	 proc{LessCrossingLines A B}
	    BC AC={CountCrossingLines Os}
	 in
	    {Swap A B}
	    BC={CountCrossingLines Os}
	    if BC>=AC then
	       {Swap A B}
	    else
	       {Assign Repeat true}
	    end
	 end
	 proc{Swap A B}
	    XYA={A getXY($)} AngA={A getAngle($)}
	    XYB={B getXY($)} AngB={B getAngle($)}
	 in
	    {A setAngle(AngB)} {A moveTo(x:XYB.1 y:XYB.2)}
	    {B setAngle(AngA)} {B moveTo(x:XYA.1 y:XYA.2)}
	 end
	 proc{MinimizeOverlap Ls}
	    {Assign Repeat false}
	    {List.forAllTail Ls
	     proc{$ Ys}
		case Ys of A|As then
		   {List.forAllTail As
		    proc{$ Xs}
		       case Xs of B|_ then
			  if {BenefitFromSwap A B} then
			     {Swap A B}
			     {Assign Repeat true}
			  end
		       else skip end
		    end}
		else skip end end}
	 end
	 proc{MinimizeCrossings Ls}
	    {Assign Repeat false}
	    {List.forAllTail Ls
	     proc{$ Ys}
		case Ys of A|As then
		   {List.forAllTail As
		    proc{$ Xs}
		       case Xs of B|_ then
			  {LessCrossingLines A B}
		       else skip end
		    end}
		else skip end end}
	 end
	 proc{While P Do}
	    if {P} then {Do} {While P Do} end
	 end
	 C=200.0
	 S=PIPI/{Int.toFloat {Length Os2}}
      in
	 %% Move all objects in a circle
	 {List.forAllInd Os2 proc{$ I X}
				A={Int.toFloat I-1}*S
				Ay={Float.toInt {Sin A}*C}
				Ax={Float.toInt {Cos A}*C}
			     in
				{X moveTo(x:Ax y:Ay)}
				{X setAngle(A)}
			     end}
	 {SetScrollRegion}
%	 {Show cl0({CountCrossingLines Os2})}
	 {While fun{$} {Access Repeat}==true end proc{$} {Show minimizeOverlap} {MinimizeOverlap Os2} end}
%	 {Show cl1({CountCrossingLines Os2})}
	 {Assign Repeat true}
	 {While fun{$} {Access Repeat}==true end proc{$} {Show minimizeCrossings} {MinimizeCrossings Os2} end}

%	 {Show cl({CountCrossingLines Os2})}
	 {ForAll Os1 proc{$ X} {X placeNearImpBy(C)} end}
      end
      %% Connect all import nodes
      {ForAll Os proc{$ N} {N drawConnections} end}
      
      %% Configure scrollregion
      {SetScrollRegion}
%      {Show done}
      {Wait DONE}
   end
in
   skip
end

/*

declare
[A]={Module.link ['FunctorMap.ozf']}
%[A]={Module.apply [AA]}
{A.display {Pickle.load "f.ozp"}}

*/



