%%%
%%% Authors:
%%%   Nils Franzen (nils.franzen@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzen, 1999
%%%   Andreas Sundstrom, 1999
%%%
%%% Last change:
%%%   $Date: 2000/03/21 13:48:22 $ by $Author: nilsf $
%%%   $Revision: 1.19 $
%%%
%%% See the file "LICENSE" or
%%%    http://mozart.ps.uni-sb.de/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
% The main program that generates html pages for the analyzed functors. 
*/

functor
import
   Pickle
   Open
   System(show:Show printInfo showInfo)
   HTML
   FunctorMap
   AnalyzeFunctor 
%   Browser(browse:Browse)
   ClassHierarchy
   ResolveFile(expandFileName:ExpandFileName getAbsolutePaths:GetAbsolutePaths)
   Comments(analyzeComment:AnalyzeComment
	    mc:MC
	    fixComments:FixComments
	    analyzeCommentCVS:AnalyzeCommentCVS
	    %%displayInfo:DisplayInfo
	    %displayExtArgs:DisplayExtArgs
	   )
   Application
   OS
export
   Analyze
define
   OZH_BASE_URL = 'x-ozlib://mozart/ozh/'
   KeepArgs
   UseCVS

   %% Parse a functor file
   fun{ParseFile FileName MacroL}
      fun {SwitchNames _} false end
      Reporter=Show
      Macros={Dictionary.new}
   in
      {ForAll MacroL proc{$ M} {Dictionary.put Macros M true} end}
      try
	 {AnalyzeFunctor.parseFile FileName Reporter SwitchNames Macros}
      catch analyze_error('No functor is found') then
	 raise
	    analyze_error('No functor is found in '#
			  {VirtualString.toAtom FileName})
	 end
      end
   end
   
   local
      %% Internal DB for keeping ref list
      RefDB={NewDictionary} IdC={NewCell 1}
      proc{NewId O} N in {Exchange IdC O N} N=O+1 end
   in
      proc{BuildDB A}
	 {Record.forAllInd A.refs
	  proc{$ Key X}
	     {Dictionary.put RefDB Key entry(X 'functor':A hrefid:{VSS "#entity"#{NewId}})}
	  end}
      end
%      fun{GetAllRefs} {Map {Dictionary.entries RefDB} fun{$ X} X.1#X.2.1.1 end} end
%      fun{GetRef K} {Dictionary.get RefDB K}.1 end
      fun{GetLinkRef K} {Dictionary.get RefDB K}.hrefid end
%      fun{GetFunctorRef K} {Dictionary.get RefDB K}.'functor' end
%      proc{ClearRefDB} {Dictionary.removeAll RefDB} end
   end
   
   %% Internal DB for keeping loaded functors
   local FunctorDB={NewDictionary} in
      proc{AddFunctor FileName Functor} {Dictionary.put FunctorDB FileName Functor} end
      fun{GetFunctor FileName} {Dictionary.get FunctorDB FileName} end
      fun{IsLoaded FileName} {Dictionary.member FunctorDB FileName} end
%      proc{ClearFunctDB} {Dictionary.removeAll FunctorDB} end
      fun{DumpFunctorDB} {Dictionary.entries FunctorDB} end
   end


   %% Convert atoms to strings to prevent that they (atoms) are mistaken for som html keywords
   A2S=Atom.toString
   VSS=VirtualString.toString
   S2A=String.toAtom
   
   %% Make records '#'(...) of lists [...]
   fun{MMap Ls F}
      {List.toRecord '#' {List.mapInd Ls fun{$ I X} I#{F X} end}}
   end
   
   %% Make anchor name
   fun{MakeLinkTag N} {VSS "<a name=\""#N.2#"\">"} /* skip initial '#' */ end

   %% Make nice header
   fun{MakeHeader Str}
      '#'(h3(Str) "\n") %'#'(font(size:"3" [Str.1]) font(size:"2" {Map Str.2 Char.toUpper}) br "\n")
   end

   %% Display a proc
   fun{DisplayProc P Functor Type}
      ID=P.id
      TypeDB={NewDictionary}
      Arity={Length P.args}
      Cm={AnalyzeComment P.comment}
   in
      {ForAll Cm.argtypes proc{$ X} {Dictionary.put TypeDB X.1 X.2} end}
      '#'("\n"
	  %%p('class':margin a({A2S P.name})) "\n"
	  code("\n" {MakeLinkTag {GetLinkRef ID}} "\n"
	       span('class':"keyword" {A2S Type}) "{" span('class':"functionname" {A2S P.name}) " "
	       {MMap {List.mapInd P.args fun{$ I X} I#X end}
		fun{$ IX}
		   I#Arg=IX
		   AT=if KeepArgs==false then "_" else {Dictionary.condGet TypeDB {S2A {Int.toString I}} Arg} end
		   S=if I==Arity then nil else " " end
		in
		   {VSS AT#S}
		end} "}"
	       if Cm.string\=nil then '#'(br "\n" span('class':"string" {FixComments Cm.string})) else nil end 
	       "\n"
	      )
	  br "\n")
   end

   %% Find an imported entity from another functor
   fun{FindImportNode N Functor}
      try
	 case {Filter {Append Functor.'import' Functor.'require'} fun{$ X} X.id==N.'import' end} 
	 of [IRef] then
	    IF={GetFunctor {S2A IRef.filename}}
	 in
	    case {Filter IF.'export'
		  fun{$ Key}
		     E=IF.refs.Key
		     Tmp={A2S (IF.refs.(E.ref)).name}
		     Name={S2A {Char.toLower Tmp.1}|Tmp.2}
		  in
		     if N.field==E.name orelse N.field==Name then true
		     else false
		     end
		  end}
	    of [ExportedEntityID] then E=IF.refs.ExportedEntityID in
	       {AdjoinAt
		{AdjoinAt
		 IF.refs.(E.ref) name if N.field\=unit then {S2A {VSS IRef.name#"."#N.field}} else IRef.name end}
		href {VSS {GetSaveFile IF.fileid}#{GetLinkRef E.ref}}}
	    [] nil then
	       nil
	    end
	 end
      catch _ then 
	 nil
      end
   end
   
   local
      CID={NewCell 0}
      Cs={NewCell nil}
      DB={NewDictionary}
   in
      proc{AddResolvedRef ID R}
	 {Dictionary.put DB ID R|{Dictionary.condGet DB ID nil}}
      end
      fun{AddClassForLaterProcess C} O N OC NC in
	 {Exchange Cs O N}
	 N=(NC#C)|O
	 {Exchange CID OC NC}
	 NC=OC+1
      end
      fun{GetAllClasses} {Access Cs}#DB end
   end

   %% Display a class
   fun{DisplayClass C Functor}
      ID=C.id
      ClassNodeId={AddClassForLaterProcess C}
      TypeDB={NewDictionary}

      fun{DisplayArg I N} 
	 case N of A1#Arg then
	    A=if {IsInt A1} then
		 {VirtualString.toAtom A1#nil}
	      elseif {IsName A1} then
		 {VirtualString.toAtom "<N:"#I#">"}
	      else A1 end
	 in
	    {VSS
	     %% Feat...
	     if I==A1 then nil else A#":" end#
	     %%  followed by value
	     if KeepArgs==false then "_" else {Dictionary.condGet TypeDB A Arg} end}
	 end
      end
      
      fun{GetArgs M S Cm}
	 Index={NewCell 1}
      in
	 {Dictionary.removeAll TypeDB}
	 {ForAll Cm.argtypes proc{$ X} {Dictionary.put TypeDB X.1 X.2} end}
	 {ForAll Cm.extargtypes proc{$ X} {Dictionary.put TypeDB X.2 X.3} end}
	 {List.foldL S fun{$ O N} OI NI in
			  {Exchange Index OI NI}
			  NI=OI+1
			  if O==nil then {DisplayArg OI N}
			  else O#" "#{DisplayArg OI N} end
		       end nil}#
	 if M==open then
	    if S\=nil then "&nbsp;..." else "..." end
	 else nil end
      end
   in
      '#'("\n" 
	  %p('class':margin a(href:{VSS "class"#ClassNodeId#".html"} {A2S C.name})) "\n"
	  {MakeLinkTag {GetLinkRef ID}} "\n"
	  code("\n"
	       %% Comments for the class
	       {MC {AnalyzeComment C.comment}}

	       %% Display class name and inheritance
	       span('class':"keyword" "class") " "
	       a(href:{VSS "class"#ClassNodeId#".html"} span('class':"type" {A2S C.name}))
	       if C.inherits\=nil then
		  '#'("&nbsp;" span('class':"keyword" "from") "&nbsp;"
		      {FoldR C.inherits
		       fun{$ K O} N=Functor.refs.K in
			  case {Label N} of 'class' then
			     {AddResolvedRef ClassNodeId K#{AdjoinAt N href {VSS {GetSaveFile Functor.fileid}#{GetLinkRef N.id}}}}
			     a(href:{GetLinkRef N.id} {A2S N.name})
			  elseof importref then
			     I={FindImportNode N Functor}
			     EName={VSS (Functor.refs.(N.'import').name)#"."#N.field}
			  in
			     if I==nil then
				{AddResolvedRef ClassNodeId K#unknown(name:{S2A EName})}
				span('class':"reference" EName)
			     else
				{AddResolvedRef ClassNodeId K#I}
				span('class':"reference" a(href:I.href EName))
			     end
			  else
			     {AddResolvedRef ClassNodeId K#unknown(name:N.name)}
			     span('class':"reference" {VSS "&lt;Unknown:"#N.name#"&gt;"})
			     %raise 'Display Inheritance'(N) end
			  end#
			  if O==nil then nil else "&nbsp;"#O end
		       end nil}
		     )
	       else nil
	       end

	       "\n" % Display all features
	       if C.features\=nil then
		  '#'(br "&nbsp;&nbsp;&nbsp;" span('class':"keyword" "feat")
		     {MMap C.features fun{$ F}
					 {VSS "&nbsp;"#F}
				      end}
		     "\n")
	       else nil end
	       
	       "\n" % Display all methods
	       ul({MMap C.methods fun{$ X}
				     Cm={AnalyzeComment X.comment}
				     Args={Record.toListInd X.args}
				  in
				     li('#'(span('class':"functionname" {A2S X.name})
					    "(" {GetArgs {Label X.args} Args Cm} ")" br "\n"
					    nil /* {MC Cm} */)) % No comments on methods at this level
				  end})

	       "\n" % End of class
	       span('class':"keyword" "end") "\n" )
	  br "\n")
   end

   fun{DisplayImport A Type}
      fun{DisplayImport2 X}
	 case X of F#V then {VSS F#":"#V} 
	 else {A2S X}
	 end
      end
   in
      code(ul({MMap A.Type
	       fun{$ X}
		  li(%% Functor Name
		     if X.filename==nil then X.name
		     else F={GetSaveFile {GetFunctor {S2A X.filename}}.fileid} in a(href:{VSS F} X.name)
		     end
		     
		     %% Imported fields
		     if X.fields==nil then nil
		     else
			"("#{FoldR X.fields
			     fun{$ N O}
				if O==nil then {DisplayImport2 N}
				else {DisplayImport2 N}#" "#O end
			     end nil}#")"
		     end
		     
		     %% 'at'
		     if X.'at'==nil then nil
		     else '#'(" at \"" span('class':"string" {VSS X.'at'} "\"\n"))
		     end
		    )
	       end}))
   end
   
   fun{DisplayExport A}
      code(ul({MMap A.'export' fun{$ Key}
				  E=A.refs.Key  % Export record
				  R=A.refs.(E.ref) % Exported entity
				  ExportFeature=if E.name==unit then nil else {VSS E.name#":"} end
			       in
				  case {Label R}
				  of 'class' then
				     li('#'(ExportFeature
					    "&lt;C:"
					    a(href:{GetLinkRef R.id}
					      span('class':"type" {A2S R.name})) "&gt;\n"))
				  [] procedure then Arity={Length R.args} in
				     li('#'(ExportFeature
					    "&lt;P/"#Arity#":"
					    a(href:{GetLinkRef R.id}
					      span('class':"functionname" {A2S R.name})) "&gt;\n"))
				  [] function then Arity={Length R.args}+1 in
				     li('#'(ExportFeature
					    "&lt;P/"#Arity#":"
					    a(href:{GetLinkRef R.id}
					      span('class':"functionname" {A2S R.name})) "&gt;\n"))
				  [] 'functor' then
				     li('#'(ExportFeature
					    "&lt;F:"
					    a(href:{GetLinkRef E.ref} % R.id} % Change as soon as Andreas fixed this
					      span('class':"type" {A2S R.name})) "&gt;\n"))
				  [] unknown then
				     li('#'(ExportFeature span('class':"variablename" {A2S R.name})) "\n")
				  end
			       end}))
   end
   
   fun{DisplayFunctor F L}
      A=F
      fun{ShowEntities Key}
	 Y=F.refs.Key
      in
	 case {Label Y}
	 of 'class' then {DisplayClass Y F}
	 [] procedure then {DisplayProc Y F 'proc'}
	 [] function then {DisplayProc Y F 'fun'}
	 [] 'functor' then
	    fun{Expand I} X=Y.refs.I in
	       %'import'([84 101 115 116 50 46 111 122] id:<N> name:'Test2')
	       {Adjoin 'import'('at':nil fields:nil filename:nil) X}
	    end
	    F1={Adjoin {Adjoin F Y} 'functor'('import':{Map Y.'import' Expand}
					      'require':{Map Y.'require' Expand})}
	 in
	    {BuildDB F1}
	    ul({DisplayFunctor F1 inline})
	 [] unknown then
	    nil
	 else
	    {Show '** ShowEntities'(Y)}
	    em(strong("Error"))
	 end
      end
   in
      '#'(
	 if L\=top andthen {HasFeature F ref} andthen {HasFeature F name} then 
	    '#'(%p('class':margin a({A2S F.name})) "\n"
		{MakeLinkTag {GetLinkRef F.ref}} "\n")
	 else nil end

	 %% Functor comment
	 span('class':"string" "\n" {MC {AnalyzeCommentCVS A.comment UseCVS}})
	 "\n"

	 %% Functor name
	 if L==top then {MakeHeader "Functor"} else '#'(span('class':"keyword" "functor") "\n" br "\n") end
	 ul({A2S F.name} 
	    if F.file==nil then br else
	       '#'(" (\"" span('class':"string" {A2S F.file}) "\")" br) 
	    end)
	 "\n"

	 if A.'require'==nil then nil else
	    '#'(if L==top then {MakeHeader "Require"} else '#'(span('class':"keyword" "require") "\n" br "\n") end
		{DisplayImport A 'require'}
		"\n")
	 end

	 %% Import
	 if A.'import'\=nil then
	    '#'(if L==top then {MakeHeader "Import"} else '#'(span('class':"keyword" "import") "\n" br "\n") end
		{DisplayImport A 'import'} "\n")
	 else nil end
	  
	 %% Export
	 if A.'export'\=nil then 
	    '#'(if L==top then {MakeHeader "Export"} else '#'(span('class':"keyword" "export") "\n" br "\n") end
		{DisplayExport A} "\n")
	 else nil end

	 %% Prepare
	 if A.'prepare'\=nil then
	    '#'(if L==top then {MakeHeader "Prepare"} else '#'(span('class':"keyword" "prepare") "\n" br "\n") end
		{MMap A.'prepare' ShowEntities})
	 else nil end

	 %% Define
	 if A.'define'\=nil then
	    '#'(if L==top then {MakeHeader "Define"} else '#'(span('class':"keyword" "define") "\n" br "\n") end
		{MMap A.'define' ShowEntities})
	 else nil end
	  
	 if L==top then {MakeHeader "End"} else '#'(span('class':"keyword" "end") br "\n") end
	 )
   end

   proc{MakeHTML IndexL Stylesheet F Body} B in
      B=body("\n"
	     p('class':"margin"
	       table(border:"0" cellspacing:"0" cellpadding:"0" "\n"
		     tr(td(width:"100" hr(width:"100"))) "\n"
		     {Record.map IndexL fun{$ X} '#'(tr(td(width:"100" X)) "\n") end}
		     tr(td(width:"100" hr(width:"100"))) "\n"
		    ))
	     "\n"
	     {DisplayFunctor F top}
	    )
      Body=html("\n "
		head("\n  " title(F.name) "\n  "
		     %%link(href:"http://www.sics.se/mozart/stylesheets/doc.css" rel:"stylesheet" type:"text/css")
		     link(href:Stylesheet rel:"stylesheet" type:"text/css")
		     "\n ")
		"\n" B
		"\n")
   end

   fun{GetSaveFile I} node#I#".html" end

   proc{DoHTML Is F OutDir Stylesheet}
      A={MakeHTML Is Stylesheet F}
   in
      {System.showInfo "Writing "#OutDir#{GetSaveFile F.fileid}}
      {HTML.out OutDir#{GetSaveFile F.fileid} A}
   end
   
   
   local
      fun{StripAtomMarks F1}
	 if F1.1==&' andthen {List.last F1}==&' then {Reverse {Reverse F1.2}.2}
	 else F1
	 end
      end
      
      fun{MakeOzFile F}
	 case {Reverse F} of &f|&z|&o|&.|As then {VSS {Reverse As}#".oz"} else F end
      end

      fun{AddFileNames FileName Functor Ls PathL}
	 {Map Ls fun{$ Key} X=Functor.refs.Key in
		    case X of 'import'(fields:Fs id:ID name:Name url:Url1) then
		       Url=if Url1==nil then nil else {StripAtomMarks {A2S Url1}} end
		    in
		       'import'(id:ID
				fields:Fs
				name:Name
				'at':Url
				filename:if Url==nil then
					    {ExpandFileName {VSS Name#".oz"} {A2S FileName} PathL}
					 else
					    {ExpandFileName {MakeOzFile Url} {A2S FileName} PathL}
					 end)
		    else raise resolveImportError(X) end
		    end
		 end}
      end
   in
      %% Load and parse a functor file, resolve all imports and require's and call BuildTrees recursively
      fun{BuildTrees Func FileName MacroL PathL ExtraRoots}
	 {System.showInfo "Loading functor "#FileName}
	 Tree1={ParseFile {A2S FileName} MacroL}
	 Imp={AddFileNames FileName Tree1 Tree1.'import' PathL}
	 Req={AddFileNames FileName Tree1 Tree1.'require' PathL}
	 Tree={AdjoinAt {AdjoinAt Tree1 'import' Imp} 'require' Req}
	 Follows={Append {Append Req Imp} ExtraRoots}
	 
      in
	 %% Needed to determine whether functor is loaded or not, functor will be inserted later again with more info
	 {AddFunctor FileName Tree}
	 
	 ((Func#FileName)#Tree)|
	 {Filter {Map Follows fun{$ F} FN={S2A F.filename} in
				 if F.filename==nil orelse {IsLoaded FN} then nil
				 else {BuildTrees F.name FN MacroL PathL nil} end
			      end} fun{$ X} X\=nil end}
      end
   end
   
   proc{Analyze Args}
      %% Guess the name for the first functor (by capitalizing the first letter in the filename
      fun{GuessFuctorName FN}
	 {List.takeWhile {Reverse {List.takeWhile {Reverse FN} fun{$ X} X\=&/ andthen X\=&\\ end}} fun{$ X} X\=&. end}
      end

      %% Create index.html file
      proc{WriteIndexFile Is1 Fs Cs1 OutDir Stylesheet}
	 Is={Record.map Is1 fun{$ X} li(X) end}
	 Cs={MMap {Sort Cs1.1 fun{$ N O} N.2.name<O.2.name end} fun{$ C}
								li(a(href:"class"#C.1#".html"
								     span('class':"type" {A2S C.2.name})))
							     end}
	 B=body("\n" h1("functor index") 
		"\n" ul({AdjoinAt Is 1 nil}) % Remove the link to this page
		"\n" h1("class index")
		"\n" ul(Cs) "\n")
	 Body=html("\n "
		   head("\n  " title("Index of functors") "\n  "
			link(href:Stylesheet rel:"stylesheet" type:"text/css")
			"\n ")
		   "\n" B "\n")
      in
	 {System.showInfo "Writing "#OutDir#"index.html"}
	 {HTML.out OutDir#"index.html" Body}
      end
      AllFiles={Filter Args.1 fun{$ X}
				 if {OS.stat X}.type==reg then
				    true
				 else
				    {System.showInfo "Dropping file: "#X#" (is not a regular file)"}
				    false
				 end
			      end}
      if AllFiles==nil then
	 {System.showInfo "** No files left to parse! ** "}
	 {Application.exit 1}
      end
      
      Stylesheet=if Args.stylesheet==nil then "ozdoc.css" else Args.stylesheet end

      FN=AllFiles.1
      ExtraRoots={Map AllFiles.2 fun{$ X} root(filename:X name:{S2A {GuessFuctorName X}}) end}
      OutDir=Args.out
      Fs={List.mapInd
	  {Flatten {BuildTrees
		    {S2A {GuessFuctorName FN}}
		    {S2A FN}
		    Args.'define'
		    {GetAbsolutePaths Args.path}
		    ExtraRoots
		   }}
	  fun{$ I X} F={Adjoin X.2 'functor'(name:X.1.1 file:X.1.2 fileid:I)} in
	     %% Overwrite previous definition in the DB with this one
	     {AddFunctor F.file F}
	     %% Add all references to a db
	     {BuildDB F}
	     I#F
	  end}
      IIs={Map Fs fun{$ F} '#'(a(href:{GetSaveFile F.1} {A2S (F.2).name}) "\n") end}
      Is={MMap if Args.index then '#'(a(href:"index.html" "Index") "\n")|IIs
	       else IIs end
	  fun{$ X} X end}
   in
      {System.showInfo "Analyzed "#{Length Fs}#" functors"}

      if Args.dump\=nil then
	 {System.showInfo "Saving functor information to \""#Args.dump#"\"!"}
	 {Pickle.save {DumpFunctorDB} Args.dump}
      end

      if Args.map then
	 %% Display graphical overview of the relation between functors
	 {FunctorMap.display
	  {Map {DumpFunctorDB} fun{$ X} F=X.2 in
				  'functor'(name:F.name
					    'import':F.'import'
					    'require':F.'require'
					    file:F.file)
			       end}}
      else
	 %% Generate html for the functors
	 UseCVS=Args.cvs
	 KeepArgs=Args.args
	 {ForAll Fs proc{$ X} {DoHTML Is X.2 OutDir Stylesheet} end}
	 
	 %% Generate html for classes
	 if Args.classes then
	    {ClassHierarchy.analyzeClasses Is OutDir Stylesheet {GetAllClasses}}
	 end
	 
	 if Args.index then
	    %% Write index file
	    {WriteIndexFile Is Fs {GetAllClasses} OutDir Stylesheet}
	 end

	 %% Copy stylesheet?
	 if Args.stylesheet==nil andthen Args.copy then
	    {ForAll ['ozdoc.css' 'page.gif'] % Thanks Denys!
	     proc{$ File}
		try
		   File_URL = OZH_BASE_URL#File
		   File_IN  = {New Open.file init(url:File_URL flags:[read])}
		   File_OUT = {New Open.file init(name:OutDir#File
						  flags:[write create truncate])}
		in
		   {File_OUT write(vs:{File_IN read(list:$ size:all)})}
		   {File_IN close}
		   {File_OUT close}
		   {System.showInfo "Copying "#File_URL#" to "#OutDir#File}
		catch X then
		   {Show X}
		end
	     end}
	 end
      end

%      {Wait _}
   end
in
   skip
end
