%%%
%%% Authors:
%%%   Nils Franzen (nils.franzen@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzen, 1999
%%%   Andreas Sundstrom, 1999
%%%
%%% Last change:
%%%   $Date: 1999/12/09 10:36:22 $ by $Author: nilsf $
%%%   $Revision: 1.4 $
%%%
%%% See the file "LICENSE" or
%%%    http://mozart.ps.uni-sb.de/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
% Determines the the full path to an functor
*/

functor 
import
   Open
   URL
   OS
export
   ExpandFileName
   GetAbsolutePaths
define
   fun{FileExists FileName}
      try  F={New Open.file init(name:FileName flags:[read])} in
	 {F close()}
	 true
      catch _ then false end
   end

   fun{GetStraightPath P}
      P1={URL.make P}
      P2={URL.normalizePath P1.path.1}
      P3={AdjoinAt P1.path 1 P2}
      P4={AdjoinAt P1 path P3}
   in
      {URL.toString P4}
   end

   local
      CWD={URL.toBase {OS.getCWD}}
   in
      fun{GetAbsolutePaths Ps}
	 {Map Ps fun{$ P} U={URL.toBase P} in
		    if {URL.isRelative U} then %% Resolve the path relative from current working directory
		       {GetStraightPath {URL.resolve CWD U}}
		    else
		       {GetStraightPath U}
		    end
		 end} 
      end
   end
   
   fun{Search FN Ps} U={URL.make FN} in
      if {URL.isAbsolute U} then
	 if {FileExists FN} then {GetStraightPath U} else nil end
      else
	 try
	    {ForAll Ps proc{$ P} P1={GetStraightPath {URL.resolve P FN}} in
			  if {FileExists P1} then raise done(P1) end end
		       end}
	    nil
	 catch done(F) then
	    F
	 end
      end
   end

   fun {GetDirName FileName}
      {Reverse {List.dropWhile {Reverse FileName} fun {$ C} C\=&/ andthen C\=&\\ end}}
   end
   
   fun{ExpandFileName FileName CurrentFileName AlternativePaths}
      {Search FileName {GetDirName CurrentFileName}|AlternativePaths}
   end
end
